@@ -1,41 +1,18 @@
-/* Copyright (C) 2005-2011, Thorvald Natvig <thorvald@natvig.com>
+// Copyright 2005-2019 The Mumble Developers. All rights reserved.
+// Use of this source code is governed by a BSD-style license
+// that can be found in the LICENSE file at the root of the
+// Mumble source tree or at <https://www.mumble.info/LICENSE>.
 
-   All rights reserved.
+#ifndef MUMBLE_MUMBLE_SERVERHANDLER_H_
+#define MUMBLE_MUMBLE_SERVERHANDLER_H_
 
-   Redistribution and use in source and binary forms, with or without
-   modification, are permitted provided that the following conditions
-   are met:
-
-   - Redistributions of source code must retain the above copyright notice,
-     this list of conditions and the following disclaimer.
-   - Redistributions in binary form must reproduce the above copyright notice,
-     this list of conditions and the following disclaimer in the documentation
-     and/or other materials provided with the distribution.
-   - Neither the name of the Mumble Developers nor the names of its
-     contributors may be used to endorse or promote products derived from this
-     software without specific prior written permission.
-
-   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
-   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-*/
-
-#ifndef SERVERHANDLER_H_
-#define SERVERHANDLER_H_
-
-#include <boost/accumulators/accumulators.hpp>
-#include <boost/accumulators/statistics/stats.hpp>
-#include <boost/accumulators/statistics/mean.hpp>
-#include <boost/accumulators/statistics/variance.hpp>
-#include <boost/shared_ptr.hpp>
+#ifndef Q_MOC_RUN
+# include <boost/accumulators/accumulators.hpp>
+# include <boost/accumulators/statistics/stats.hpp>
+# include <boost/accumulators/statistics/mean.hpp>
+# include <boost/accumulators/statistics/variance.hpp>
+# include <boost/shared_ptr.hpp>
+#endif
 
 #include <QtCore/QEvent>
 #include <QtCore/QMutex>
@@ -55,11 +32,14 @@
 #include "Timer.h"
 #include "Message.h"
 #include "Mumble.pb.h"
+#include "ServerAddress.h"
 
 class Connection;
+class Database;
 class Message;
 class PacketDataStream;
 class QUdpSocket;
+class QSslSocket;
 class VoiceRecorder;
 
 class ServerHandlerMessageEvent : public QEvent {
@@ -76,11 +56,14 @@ class ServerHandler : public QThread {
 	private:
 		Q_OBJECT
 		Q_DISABLE_COPY(ServerHandler)
+
+		Database *database;
 	protected:
 		QString qsHostName;
 		QString qsUserName;
 		QString qsPassword;
 		unsigned short usPort;
+		unsigned short usResolvedPort;
 		bool bUdp;
 		bool bStrong;
 
@@ -90,12 +73,14 @@ class ServerHandler : public QThread {
 #endif
 
 		QHostAddress qhaRemote;
+		QHostAddress qhaLocal;
 		QUdpSocket *qusUdp;
 		QMutex qmUdp;
 
 		void handleVoicePacket(unsigned int msgFlags, PacketDataStream &pds, MessageHandler::UDPMessageType type);
 	public:
 		Timer tTimestamp;
+		int iInFlightTCPPings;
 		QTimer *tConnectionTimeoutTimer;
 		QList<QSslError> qlErrors;
 		QList<QSslCertificate> qscCert;
@@ -103,6 +88,9 @@ class ServerHandler : public QThread {
 		ConnectionPtr cConnection;
 		QByteArray qbaDigest;
 		boost::shared_ptr<VoiceRecorder> recorder;
+		QSslSocket *qtsSock;
+		QList<ServerAddress> qlAddresses;
+		ServerAddress saTargetServer;
 
 		unsigned int uiVersion;
 		QString qsRelease;
@@ -116,7 +104,7 @@ class ServerHandler : public QThread {
 		void setConnectionInfo(const QString &host, unsigned short port, const QString &username, const QString &pw);
 		void getConnectionInfo(QString &host, unsigned short &port, QString &username, QString &pw) const;
 		bool isStrong() const;
-		void customEvent(QEvent *evt);
+		void customEvent(QEvent *evt) Q_DECL_OVERRIDE;
 
 		void sendProtoMessage(const ::google::protobuf::Message &msg, unsigned int msgType);
 		void sendMessage(const char *data, int len, bool force = false);
@@ -126,9 +114,8 @@ class ServerHandler : public QThread {
 #undef MUMBLE_MH_MSG
 
 		void requestUserStats(unsigned int uiSession, bool statsOnly);
-		void joinChannel(unsigned int channel);
-		void createChannel(unsigned int parent_, const QString &name, const QString &description, unsigned int position, bool temporary);
-		void setTexture(const QByteArray &qba);
+		void joinChannel(unsigned int uiSession, unsigned int channel);
+		void createChannel(unsigned int parent_id, const QString &name, const QString &description, unsigned int position, bool temporary, unsigned int maxUsers);
 		void requestBanList();
 		void requestUserList();
 		void requestACL(unsigned int channel);
@@ -137,18 +124,25 @@ class ServerHandler : public QThread {
 		void sendUserTextMessage(unsigned int uiSession, const QString &message_);
 		void sendChannelTextMessage(unsigned int channel, const QString &message_, bool tree);
 		void setUserComment(unsigned int uiSession, const QString &comment);
+		void setUserTexture(unsigned int uiSession, const QByteArray &qba);
+		void setTokens(const QStringList &tokens);
 		void removeChannel(unsigned int channel);
 		void addChannelLink(unsigned int channel, unsigned int link);
 		void removeChannelLink(unsigned int channel, unsigned int link);
 		void requestChannelPermissions(unsigned int channel);
 		void setSelfMuteDeafState(bool mute, bool deaf);
 		void announceRecordingState(bool recording);
+		
+		/// Return connection information as a URL
+		QUrl getServerURL(bool withPassword = false) const;
 
 		void disconnect();
-		void run();
+		void run() Q_DECL_OVERRIDE;
 	signals:
+		void error(QAbstractSocket::SocketError, QString reason);
 		void disconnected(QAbstractSocket::SocketError, QString reason);
 		void connected();
+		void pingRequested();
 	protected slots:
 		void message(unsigned int, const QByteArray &);
 		void serverConnectionConnected();
@@ -157,6 +151,9 @@ class ServerHandler : public QThread {
 		void serverConnectionClosed(QAbstractSocket::SocketError, const QString &);
 		void setSslErrors(const QList<QSslError> &);
 		void udpReady();
+		void hostnameResolved();
+	private slots:
+		void sendPingInternal();
 	public slots:
 		void sendPing();
 };
